#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Compute v2 Server action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import importlib
from typing import Any

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


class CreateServerBackup(command.ShowOne):
    _description = _("Create a server backup image")

    IMAGE_API_VERSIONS = {
        "1": "openstackclient.image.v1.image",
        "2": "openstackclient.image.v2.image",
    }

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server to back up (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<image-name>',
            help=_('Name of the backup image (default: server name)'),
        )
        parser.add_argument(
            '--type',
            metavar='<backup-type>',
            help=_(
                'Used to populate the backup_type property of the backup '
                'image (default: empty)'
            ),
        )
        parser.add_argument(
            '--rotate',
            metavar='<count>',
            type=int,
            help=_('Number of backups to keep (default: 1)'),
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for backup image create to complete'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        def _show_progress(progress: int | None) -> None:
            if progress:
                self.app.stderr.write(f'\rProgress: {progress}')
                self.app.stderr.flush()

        compute_client = self.app.client_manager.compute

        server = compute_client.find_server(
            parsed_args.server, ignore_missing=False
        )

        # Set sane defaults as this API wants all mouths to be fed
        if parsed_args.name is None:
            backup_name = server.name
        else:
            backup_name = parsed_args.name
        if parsed_args.type is None:
            backup_type = ""
        else:
            backup_type = parsed_args.type
        if parsed_args.rotate is None:
            backup_rotation = 1
        else:
            backup_rotation = parsed_args.rotate

        compute_client.backup_server(
            server.id,
            backup_name,
            backup_type,
            backup_rotation,
        )

        image_client = self.app.client_manager.image
        image = image_client.find_image(backup_name, ignore_missing=False)

        if parsed_args.wait:
            if utils.wait_for_status(
                image_client.get_image,
                image.id,
                callback=_show_progress,
            ):
                self.app.stdout.write('\n')
            else:
                msg = _('Error creating server backup: %s') % parsed_args.name
                raise exceptions.CommandError(msg)

        if self.app.client_manager._api_version['image'] == '1':
            info = {}
            info.update(image._info)
            info['properties'] = utils.format_dict(info.get('properties', {}))
        else:
            # Get the right image module to format the output
            image_module = importlib.import_module(
                self.IMAGE_API_VERSIONS[
                    self.app.client_manager._api_version['image']
                ]
            )
            info = image_module._format_image(image)
        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstackclient.network.v2 import (
    network_service_provider as service_provider,
)
from openstackclient.tests.unit.network.v2 import fakes


class TestListNetworkServiceProvider(fakes.TestNetworkV2):
    provider_list = (
        fakes.FakeNetworkServiceProvider.create_network_service_providers(
            count=2
        )
    )

    columns = (
        'Service Type',
        'Name',
        'Default',
    )

    data = []

    for provider in provider_list:
        data.append(
            (
                provider.service_type,
                provider.name,
                provider.is_default,
            )
        )

    def setUp(self):
        super().setUp()
        self.network_client.service_providers.return_value = self.provider_list

        self.cmd = service_provider.ListNetworkServiceProvider(self.app, None)

    def test_network_service_provider_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.service_providers.assert_called_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ast
import json

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v3 import common


class AccessRuleTests(common.IdentityTests):
    ACCESS_RULE_FIELDS = [
        'ID',
        'Service',
        'Method',
        'Path',
    ]
    ACCESS_RULE_LIST_HEADERS = [
        'ID',
        'Service',
        'Method',
        'Path',
    ]

    def setUp(self):
        super().setUp()

        application_credential_name = data_utils.rand_name('name')
        access_rules = json.dumps(
            [
                {
                    'method': 'GET',
                    'path': '/v2.1/servers',
                    'service': 'compute',
                },
                {
                    'method': 'GET',
                    'path': '/v2.0/networks',
                    'service': 'networking',
                },
            ]
        )
        raw_output = self.openstack(
            f"application credential create {application_credential_name} "
            f"--access-rules '{access_rules}'"
        )
        # we immediately delete the application credential since it will leave
        # the access rules around
        self.openstack(
            f'application credential delete {application_credential_name}'
        )

        items = self.parse_show_as_object(raw_output)
        self.access_rule_ids = [
            x['id'] for x in ast.literal_eval(items['Access Rules'])
        ]
        self.addCleanup(
            self.openstack,
            'access rule delete '
            + ' '.join([x for x in self.access_rule_ids]),
        )

    def test_access_rule(self):
        # list

        raw_output = self.openstack('access rule list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, self.ACCESS_RULE_LIST_HEADERS)

        # show

        raw_output = self.openstack(
            f'access rule show {self.access_rule_ids[0]}'
        )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.ACCESS_RULE_FIELDS)



# Copyright 2026 Openinfra Foundation
#
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import operator
from unittest import mock

from osc_lib import exceptions
from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from openstackclient.network.v2.bgpvpn import port_association
from openstackclient.tests.unit.network.v2.bgpvpn import fakes


columns_short = tuple(
    col
    for col, _, listing_mode in port_association._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_SHORT_ONLY)
)
columns_long = tuple(
    col
    for col, _, listing_mode in port_association._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_LONG_ONLY)
)
headers_short = tuple(
    head
    for _, head, listing_mode in port_association._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_SHORT_ONLY)
)
headers_long = tuple(
    head
    for _, head, listing_mode in port_association._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_LONG_ONLY)
)
sorted_attr_map = sorted(
    port_association._attr_map, key=operator.itemgetter(1)
)
sorted_columns = tuple(col for col, _, _ in sorted_attr_map)
sorted_headers = tuple(head for _, head, _ in sorted_attr_map)


def _get_data(attrs, columns=sorted_columns):
    return osc_utils.get_dict_properties(
        attrs, columns, formatters=port_association._formatters
    )


class TestCreatePortAssoc(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.cmd = port_association.CreateBgpvpnPortAssoc(self.app, None)
        self.fake_bgpvpn = fakes.create_one_bgpvpn()
        self.fake_port = fakes.create_one_resource()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.network_client.find_port.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.fake_project = mock.Mock(id='fake_project_id')
        self.identity_sdk_client.find_project = mock.Mock(
            return_value=self.fake_project
        )

    @mock.patch('osc_lib.cli.identity.find_project')
    def test_create_port_association(self, mock_find_project):
        fake_assoc = fakes.create_one_port_association()
        self.network_client.create_bgpvpn_port_association.return_value = (
            fake_assoc
        )
        mock_find_project.return_value = mock.Mock(
            id=self.fake_bgpvpn['project_id']
        )
        arglist = [
            self.fake_bgpvpn['id'],
            self.fake_port['id'],
            '--project',
            self.fake_bgpvpn['project_id'],
        ]
        verifylist = [
            ('bgpvpn', self.fake_bgpvpn['id']),
            ('port', self.fake_port['id']),
            ('project', self.fake_bgpvpn['project_id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        cols, _data = self.cmd.take_action(parsed_args)

        self.network_client.create_bgpvpn_port_association.assert_called_once_with(
            self.fake_bgpvpn['id'],
            port_id=self.fake_port['id'],
            project_id='fake_project_id',
            routes=[],
        )
        self.assertIn('id', cols)
        self.assertIn('port_id', cols)


class TestSetPortAssoc(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.cmd = port_association.SetBgpvpnPortAssoc(self.app, None)
        self.fake_bgpvpn = fakes.create_one_bgpvpn()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.network_client.find_bgpvpn_port_association.return_value = {
            'routes': []
        }

    def test_set_port_association_advertise(self):
        fake_assoc = fakes.create_one_port_association()
        arglist = [
            fake_assoc['id'],
            self.fake_bgpvpn['id'],
            '--advertise-fixed-ips',
        ]
        verifylist = [
            ('port_association_id', fake_assoc['id']),
            ('bgpvpn', self.fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_bgpvpn_port_association.assert_called_once_with(
            self.fake_bgpvpn['id'],
            fake_assoc['id'],
            advertise_fixed_ips=True,
            routes=[],
        )
        self.assertIsNone(result)

    def test_set_port_association_no_advertise(self):
        fake_assoc = fakes.create_one_port_association()
        arglist = [
            fake_assoc['id'],
            self.fake_bgpvpn['id'],
            '--no-advertise-fixed-ips',
        ]
        verifylist = [
            ('port_association_id', fake_assoc['id']),
            ('bgpvpn', self.fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_bgpvpn_port_association.assert_called_once_with(
            self.fake_bgpvpn['id'],
            fake_assoc['id'],
            advertise_fixed_ips=False,
            routes=[],
        )
        self.assertIsNone(result)


class TestUnsetPortAssoc(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.cmd = port_association.UnsetBgpvpnPortAssoc(self.app, None)
        self.fake_bgpvpn = fakes.create_one_bgpvpn()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.network_client.find_bgpvpn_port_association.return_value = {
            'routes': []
        }

    def test_unset_port_association_advertise(self):
        fake_assoc = fakes.create_one_port_association()
        arglist = [
            fake_assoc['id'],
            self.fake_bgpvpn['id'],
            '--advertise-fixed-ips',
        ]
        verifylist = [
            ('port_association_id', fake_assoc['id']),
            ('bgpvpn', self.fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_bgpvpn_port_association.assert_called_once_with(
            self.fake_bgpvpn['id'],
            fake_assoc['id'],
            advertise_fixed_ips=False,
            routes=[],
        )
        self.assertIsNone(result)


class TestDeletePortAssoc(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.cmd = port_association.DeleteBgpvpnPortAssoc(self.app, None)

    def test_delete_one_association(self):
        fake_bgpvpn = fakes.create_one_bgpvpn()
        fake_assoc = fakes.create_one_port_association()
        arglist = [
            fake_assoc['id'],
            fake_bgpvpn['id'],
        ]
        verifylist = [
            ('port_association_ids', [fake_assoc['id']]),
            ('bgpvpn', fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_bgpvpn_port_association.assert_called_once_with(
            fake_bgpvpn['id'], fake_assoc['id']
        )
        self.assertIsNone(result)

    def test_delete_multi_association(self):
        count = 3
        fake_bgpvpn = fakes.create_one_bgpvpn()
        fake_assocs = fakes.create_port_associations(count=count)
        fake_assoc_ids = [a['id'] for a in fake_assocs]
        arglist = [*fake_assoc_ids, fake_bgpvpn['id']]
        verifylist = [
            ('port_association_ids', fake_assoc_ids),
            ('bgpvpn', fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_bgpvpn_port_association.assert_has_calls(
            [mock.call(fake_bgpvpn['id'], id) for id in fake_assoc_ids]
        )
        self.assertIsNone(result)

    def test_delete_multi_association_with_unknown(self):
        count = 3
        fake_bgpvpn = fakes.create_one_bgpvpn()
        fake_assocs = fakes.create_port_associations(count=count)
        fake_assoc_ids = [a['id'] for a in fake_assocs]

        def raise_unknown_resource(bgpvpn_id, association_id):
            if str(count - 2) in association_id:
                raise Exception()

        self.network_client.delete_bgpvpn_port_association.side_effect = (
            raise_unknown_resource
        )
        arglist = [*fake_assoc_ids, fake_bgpvpn['id']]
        verifylist = [
            ('port_association_ids', fake_assoc_ids),
            ('bgpvpn', fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

        self.network_client.delete_bgpvpn_port_association.assert_has_calls(
            [mock.call(fake_bgpvpn['id'], id) for id in fake_assoc_ids]
        )


class TestListPortAssoc(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.cmd = port_association.ListBgpvpnPortAssoc(self.app, None)

    def test_list_port_associations(self):
        count = 3
        fake_bgpvpn = fakes.create_one_bgpvpn()
        fake_assocs = fakes.create_port_associations(count=count)
        self.network_client.bgpvpn_port_associations.return_value = fake_assocs
        arglist = [
            fake_bgpvpn['id'],
        ]
        verifylist = [
            ('bgpvpn', fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.bgpvpn_port_associations.assert_called_once_with(
            fake_bgpvpn['id'], retrieve_all=True
        )
        self.assertEqual(headers, list(headers_short))
        self.assertEqual(
            list(data),
            [_get_data(a, columns_short) for a in fake_assocs],
        )

    def test_list_port_associations_long_mode(self):
        count = 3
        fake_bgpvpn = fakes.create_one_bgpvpn()
        fake_assocs = fakes.create_port_associations(count=count)
        self.network_client.bgpvpn_port_associations.return_value = fake_assocs
        arglist = [
            '--long',
            fake_bgpvpn['id'],
        ]
        verifylist = [
            ('long', True),
            ('bgpvpn', fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.bgpvpn_port_associations.assert_called_once_with(
            fake_bgpvpn['id'], retrieve_all=True
        )
        self.assertEqual(headers, list(headers_long))
        self.assertEqual(
            list(data),
            [_get_data(a, columns_long) for a in fake_assocs],
        )


class TestShowPortAssoc(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.cmd = port_association.ShowBgpvpnPortAssoc(self.app, None)

    def test_show_port_association(self):
        fake_bgpvpn = fakes.create_one_bgpvpn()
        fake_assoc = fakes.create_one_port_association()
        self.network_client.get_bgpvpn_port_association.return_value = (
            fake_assoc
        )

        arglist = [
            fake_assoc['id'],
            fake_bgpvpn['id'],
        ]
        verifylist = [
            ('port_association_id', fake_assoc['id']),
            ('bgpvpn', fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        cols, _data = self.cmd.take_action(parsed_args)

        self.network_client.get_bgpvpn_port_association.assert_called_once_with(
            fake_bgpvpn['id'], fake_assoc['id']
        )
        # _transform_resource removes 'routes' and only adds
        # prefix_routes/bgpvpn_routes if routes exist, so with empty
        # routes the show columns won't include those fields
        self.assertIn('id', cols)
        self.assertIn('port_id', cols)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from unittest import mock

from openstackclient.network.v2.dynamic_routing import bgp_dragent
from openstackclient.tests.unit.network.v2.dynamic_routing import fakes
from openstackclient.tests.unit.network.v2 import fakes as test_fakes


class TestAddBgpSpeakerToDRAgent(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self._bgp_speaker = fakes.FakeBgpSpeaker.create_one_bgp_speaker()
        self._bgp_dragent = fakes.FakeDRAgent.create_one_dragent()
        self._bgp_speaker_id = self._bgp_speaker['id']
        self._bgp_dragent_id = self._bgp_dragent['id']

        # Get the command object to test
        self.cmd = bgp_dragent.AddBgpSpeakerToDRAgent(self.app, None)

    def test_add_bgp_speaker_to_dragent(self):
        arglist = [
            self._bgp_dragent_id,
            self._bgp_speaker_id,
        ]
        verifylist = [
            ('dragent_id', self._bgp_dragent_id),
            ('bgp_speaker', self._bgp_speaker_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.network_client.find_bgp_speaker.return_value = mock.Mock(
            id=self._bgp_speaker_id
        )
        self.network_client.add_bgp_speaker_to_dragent.return_value = None

        result = self.cmd.take_action(parsed_args)
        self.network_client.add_bgp_speaker_to_dragent.assert_called_once_with(
            self._bgp_dragent_id, self._bgp_speaker_id
        )
        self.assertIsNone(result)


class TestRemoveBgpSpeakerFromDRAgent(test_fakes.TestNetworkV2):
    _bgp_speaker = fakes.FakeBgpSpeaker.create_one_bgp_speaker()
    _bgp_dragent = fakes.FakeDRAgent.create_one_dragent()
    _bgp_speaker_id = _bgp_speaker['id']
    _bgp_dragent_id = _bgp_dragent['id']

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = bgp_dragent.RemoveBgpSpeakerFromDRAgent(self.app, None)

    def test_remove_bgp_speaker_from_dragent(self):
        arglist = [
            self._bgp_dragent_id,
            self._bgp_speaker_id,
        ]
        verifylist = [
            ('dragent_id', self._bgp_dragent_id),
            ('bgp_speaker', self._bgp_speaker_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.network_client.find_bgp_speaker.return_value = mock.Mock(
            id=self._bgp_speaker_id
        )
        self.network_client.remove_bgp_speaker_from_dragent.return_value = None

        result = self.cmd.take_action(parsed_args)
        self.network_client.remove_bgp_speaker_from_dragent.assert_called_once_with(
            self._bgp_dragent_id, self._bgp_speaker_id
        )
        self.assertIsNone(result)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstackclient.tests.functional.identity.v3 import common


class TokenTests(common.IdentityTests):
    def test_token_issue(self):
        raw_output = self.openstack('token issue')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.TOKEN_FIELDS)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import contextlib
from unittest import mock

from osc_lib import exceptions

from openstack import exceptions as sdk_exc
from openstack.identity.v3 import domain as _domain
from openstack.identity.v3 import group as _group
from openstack.identity.v3 import project as _project
from openstack.identity.v3 import role_assignment as _role_assignment
from openstack.identity.v3 import user as _user
from openstack.test import fakes as sdk_fakes

from openstackclient.identity import common
from openstackclient.identity.v3 import user
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestUserCreate(identity_fakes.TestIdentity):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)
    project = sdk_fakes.generate_fake_resource(_project.Project)

    columns = (
        'default_project_id',
        'domain_id',
        'email',
        'enabled',
        'id',
        'name',
        'description',
        'password_expires_at',
        'options',
    )

    def setUp(self):
        super().setUp()

        self.user = sdk_fakes.generate_fake_resource(
            resource_type=_user.User,
            domain_id=self.domain.id,
            default_project_id=self.project.id,
        )
        self.datalist = (
            self.project.id,
            self.domain.id,
            self.user.email,
            True,
            self.user.id,
            self.user.name,
            self.user.description,
            self.user.password_expires_at,
            getattr(self.user, 'options', {}),
        )

        self.identity_sdk_client.find_domain.return_value = self.domain
        self.identity_sdk_client.find_project.return_value = self.project
        self.identity_sdk_client.create_user.return_value = self.user

        # Get the command object to test
        self.cmd = user.CreateUser(self.app, None)

    def test_user_create_no_options(self):
        arglist = [
            self.user.name,
        ]
        verifylist = [
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_password(self):
        arglist = [
            '--password',
            'secret',
            self.user.name,
        ]
        verifylist = [
            ('password', 'secret'),
            ('password_prompt', False),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'password': 'secret',
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_password_prompt(self):
        arglist = [
            '--password-prompt',
            self.user.name,
        ]
        verifylist = [
            ('password_prompt', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        mocker = mock.Mock()
        mocker.return_value = 'abc123'
        with mock.patch("osc_lib.utils.get_password", mocker):
            columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'password': 'abc123',
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_password_prompt_no_warning(self):
        arglist = [
            '--password-prompt',
            self.user.name,
        ]
        verifylist = [
            ('password_prompt', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        import logging

        # Mock the password prompt
        mocker = mock.Mock()
        mocker.return_value = 'abc123'

        # Use assertLogs to verify no warnings are logged
        logger = 'openstackclient.identity.v3.user'
        with mock.patch("osc_lib.utils.get_password", mocker):
            with self.assertLogs(logger, level='WARNING') as log_ctx:
                logging.getLogger(logger).warning(
                    "Dummy warning for test setup"
                )
                columns, data = self.cmd.take_action(parsed_args)

                self.assertEqual(1, len(log_ctx.records))
                self.assertIn(
                    "Dummy warning for test setup", log_ctx.output[0]
                )
                self.assertNotIn(
                    "No password was supplied", ''.join(log_ctx.output)
                )

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'password': 'abc123',
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_email(self):
        arglist = [
            '--email',
            'barney@example.com',
            self.user.name,
        ]
        verifylist = [
            ('email', 'barney@example.com'),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'email': 'barney@example.com',
            'is_enabled': True,
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_project(self):
        arglist = [
            '--project',
            self.project.name,
            self.user.name,
        ]
        verifylist = [
            ('project', self.project.name),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'default_project_id': self.project.id,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            self.project.id,
            self.domain.id,
            self.user.email,
            True,
            self.user.id,
            self.user.name,
            self.user.description,
            self.user.password_expires_at,
            getattr(self.user, 'options', {}),
        )
        self.assertEqual(datalist, data)

    def test_user_create_project_domain(self):
        arglist = [
            '--project',
            self.project.name,
            '--project-domain',
            self.project.domain_id,
            self.user.name,
        ]
        verifylist = [
            ('project', self.project.name),
            ('project_domain', self.project.domain_id),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'default_project_id': self.project.id,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_user.assert_called_once_with(**kwargs)
        self.identity_sdk_client.find_domain.assert_called_once_with(
            self.project.domain_id, ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            self.project.id,
            self.domain.id,
            self.user.email,
            True,
            self.user.id,
            self.user.name,
            self.user.description,
            self.user.password_expires_at,
            getattr(self.user, 'options', {}),
        )
        self.assertEqual(datalist, data)

    def test_user_create_domain(self):
        arglist = [
            '--domain',
            self.domain.name,
            self.user.name,
        ]
        verifylist = [
            ('domain', self.domain.name),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'domain_id': self.domain.id,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_enable(self):
        arglist = [
            '--enable',
            self.user.name,
        ]
        verifylist = [
            ('enable', True),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_disable(self):
        arglist = [
            '--disable',
            self.user.name,
        ]
        verifylist = [
            ('name', self.user.name),
            ('enable', False),
            ('disable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': False,
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_ignore_lockout_failure_attempts(self):
        arglist = [
            '--ignore-lockout-failure-attempts',
            self.user.name,
        ]
        verifylist = [
            ('ignore_lockout_failure_attempts', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'ignore_lockout_failure_attempts': True},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_no_ignore_lockout_failure_attempts(self):
        arglist = [
            '--no-ignore-lockout-failure-attempts',
            self.user.name,
        ]
        verifylist = [
            ('no_ignore_lockout_failure_attempts', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'ignore_lockout_failure_attempts': False},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_ignore_password_expiry(self):
        arglist = [
            '--ignore-password-expiry',
            self.user.name,
        ]
        verifylist = [
            ('ignore_password_expiry', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'ignore_password_expiry': True},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_no_ignore_password_expiry(self):
        arglist = [
            '--no-ignore-password-expiry',
            self.user.name,
        ]
        verifylist = [
            ('no_ignore_password_expiry', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'ignore_password_expiry': False},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_ignore_change_password_upon_first_use(self):
        arglist = [
            '--ignore-change-password-upon-first-use',
            self.user.name,
        ]
        verifylist = [
            ('ignore_change_password_upon_first_use', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'ignore_change_password_upon_first_use': True},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_no_ignore_change_password_upon_first_use(self):
        arglist = [
            '--no-ignore-change-password-upon-first-use',
            self.user.name,
        ]
        verifylist = [
            ('no_ignore_change_password_upon_first_use', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'ignore_change_password_upon_first_use': False},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_enables_lock_password(self):
        arglist = [
            '--enable-lock-password',
            self.user.name,
        ]
        verifylist = [
            ('enable_lock_password', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'lock_password': True},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_disables_lock_password(self):
        arglist = [
            '--disable-lock-password',
            self.user.name,
        ]
        verifylist = [
            ('disable_lock_password', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'lock_password': False},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_enable_multi_factor_auth(self):
        arglist = [
            '--enable-multi-factor-auth',
            self.user.name,
        ]
        verifylist = [
            ('enable_multi_factor_auth', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'multi_factor_auth_enabled': True},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_disable_multi_factor_auth(self):
        arglist = [
            '--disable-multi-factor-auth',
            self.user.name,
        ]
        verifylist = [
            ('disable_multi_factor_auth', True),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {'multi_factor_auth_enabled': False},
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_option_with_multi_factor_auth_rule(self):
        arglist = [
            '--multi-factor-auth-rule',
            identity_fakes.mfa_opt1,
            '--multi-factor-auth-rule',
            identity_fakes.mfa_opt2,
            self.user.name,
        ]
        verifylist = [
            (
                'multi_factor_auth_rules',
                [identity_fakes.mfa_opt1, identity_fakes.mfa_opt2],
            ),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {
                'multi_factor_auth_rules': [["password", "totp"], ["password"]]
            },
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_with_multiple_options(self):
        arglist = [
            '--ignore-password-expiry',
            '--disable-multi-factor-auth',
            '--multi-factor-auth-rule',
            identity_fakes.mfa_opt1,
            self.user.name,
        ]
        verifylist = [
            ('ignore_password_expiry', True),
            ('disable_multi_factor_auth', True),
            ('multi_factor_auth_rules', [identity_fakes.mfa_opt1]),
            ('enable', False),
            ('disable', False),
            ('name', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.user.name,
            'is_enabled': True,
            'options': {
                'ignore_password_expiry': True,
                'multi_factor_auth_enabled': False,
                'multi_factor_auth_rules': [["password", "totp"]],
            },
        }
        self.identity_sdk_client.create_user.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)


class TestUserDelete(identity_fakes.TestIdentity):
    user = sdk_fakes.generate_fake_resource(_user.User)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.find_user.return_value = self.user
        self.identity_sdk_client.delete_user.return_value = None

        # Get the command object to test
        self.cmd = user.DeleteUser(self.app, None)

    def test_user_delete_no_options(self):
        arglist = [
            self.user.id,
        ]
        verifylist = [
            ('users', [self.user.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_user.assert_called_with(
            self.user.id,
            ignore_missing=False,
        )
        self.assertIsNone(result)

    @mock.patch.object(_user.User, 'find')
    def test_delete_multi_users_with_exception(self, find_mock):
        self.identity_sdk_client.find_user.side_effect = [
            self.user,
            sdk_exc.ResourceNotFound,
        ]
        arglist = [
            self.user.id,
            'unexist_user',
        ]
        verifylist = [
            ('users', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 users failed to delete.', str(e))

        self.identity_sdk_client.find_user.assert_has_calls(
            [
                mock.call(name_or_id=self.user.id, ignore_missing=False),
                mock.call(name_or_id='unexist_user', ignore_missing=False),
            ]
        )

        self.assertEqual(2, self.identity_sdk_client.find_user.call_count)
        self.identity_sdk_client.delete_user.assert_called_once_with(
            self.user.id, ignore_missing=False
        )


class TestUserList(identity_fakes.TestIdentity):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)
    project = sdk_fakes.generate_fake_resource(_project.Project)
    user = sdk_fakes.generate_fake_resource(
        resource_type=_user.User,
        domain_id=domain.id,
        default_project_id=project.id,
    )
    group = sdk_fakes.generate_fake_resource(_group.Group)
    role_assignment = sdk_fakes.generate_fake_resource(
        resource_type=_role_assignment.RoleAssignment, user={'id': user.id}
    )

    columns = ['ID', 'Name']
    datalist = (
        (
            user.id,
            user.name,
        ),
    )

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.find_user.return_value = self.user
        self.identity_sdk_client.users.return_value = [self.user]
        self.identity_sdk_client.group_users.return_value = [self.user]
        self.identity_sdk_client.find_domain.return_value = self.domain
        self.identity_sdk_client.find_group.return_value = self.group
        self.identity_sdk_client.find_project.return_value = self.project
        self.identity_sdk_client.role_assignments.return_value = [
            self.role_assignment
        ]

        # Get the command object to test
        self.cmd = user.ListUser(self.app, None)

    def test_user_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'domain_id': None,
        }

        self.identity_sdk_client.users.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_user_list_domain(self):
        arglist = [
            '--domain',
            self.domain.id,
        ]
        verifylist = [
            ('domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'domain_id': self.domain.id,
        }

        self.identity_sdk_client.users.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_user_list_group(self):
        arglist = [
            '--group',
            self.group.name,
        ]
        verifylist = [
            ('group', self.group.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'domain_id': None,
            'group': self.group.id,
        }

        self.identity_sdk_client.group_users.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_user_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'domain_id': None,
        }

        self.identity_sdk_client.users.assert_called_with(**kwargs)

        collist = [
            'ID',
            'Name',
            'Project',
            'Domain',
            'Description',
            'Email',
            'Enabled',
        ]
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.user.id,
                self.user.name,
                self.project.id,
                self.domain.id,
                self.user.description,
                self.user.email,
                True,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_user_list_project(self):
        arglist = [
            '--project',
            self.project.name,
        ]
        verifylist = [
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'scope_project_id': self.project.id,
        }

        self.identity_sdk_client.role_assignments.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_user_list_with_option_enabled(self):
        arglist = ['--enabled']
        verifylist = [('is_enabled', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'domain_id': None, 'is_enabled': True}
        self.identity_sdk_client.users.assert_called_with(**kwargs)
        self.identity_sdk_client.find_user.assert_not_called()
        self.identity_sdk_client.group_users.assert_not_called()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_user_list_with_pagination(self):
        arglist = [
            '--limit',
            '2',
            '--marker',
            'some-marker',
        ]
        verifylist = [
            ('limit', 2),
            ('marker', 'some-marker'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'domain_id': None,
            'limit': 2,
            'marker': 'some-marker',
        }
        self.identity_sdk_client.users.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_user_list_with_pagination_and_group(self):
        arglist = [
            '--group',
            self.group.name,
            '--limit',
            '5',
        ]
        verifylist = [
            ('group', self.group.name),
            ('limit', 5),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'domain_id': None,
            'group': self.group.id,
            'limit': 5,
        }
        self.identity_sdk_client.group_users.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_user_list_pagination_with_project_fails(self):
        arglist = [
            '--project',
            self.project.name,
            '--limit',
            '2',
        ]
        verifylist = [
            ('project', self.project.name),
            ('limit', 2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestUserSet(identity_fakes.TestIdentity):
    project = sdk_fakes.generate_fake_resource(_project.Project)
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)
    user = sdk_fakes.generate_fake_resource(
        resource_type=_user.User, default_project_id=project.id
    )
    user2 = sdk_fakes.generate_fake_resource(
        resource_type=_user.User,
        default_project_id=project.id,
        domain_id=domain.id,
    )

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.find_project.return_value = self.project
        self.identity_sdk_client.find_user.return_value = self.user
        self.identity_sdk_client.update_user.return_value = self.user

        # Get the command object to test
        self.cmd = user.SetUser(self.app, None)

    def test_user_set_no_options(self):
        arglist = [
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

    def test_user_set_name(self):
        arglist = [
            '--name',
            'qwerty',
            self.user.name,
        ]
        verifylist = [
            ('name', 'qwerty'),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
            'name': 'qwerty',
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_specify_domain(self):
        arglist = [
            '--name',
            'qwerty',
            '--domain',
            self.domain.id,
            self.user2.name,
        ]
        verifylist = [
            ('name', 'qwerty'),
            ('domain', self.domain.id),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user2.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        kwargs = {'is_enabled': True, 'name': 'qwerty'}

        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_password(self):
        arglist = [
            '--password',
            'secret',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('password', 'secret'),
            ('password_prompt', False),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
            'password': 'secret',
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_password_prompt(self):
        arglist = [
            '--password-prompt',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('password_prompt', True),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        mocker = mock.Mock()
        mocker.return_value = 'abc123'
        with mock.patch("osc_lib.utils.get_password", mocker):
            result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
            'password': 'abc123',
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_email(self):
        arglist = [
            '--email',
            'barney@example.com',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', 'barney@example.com'),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
            'email': 'barney@example.com',
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_project(self):
        arglist = [
            '--project',
            self.project.id,
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('project', self.project.id),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
            'default_project_id': self.project.id,
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.identity_sdk_client.find_domain.assert_not_called()

        # Set expected values
        kwargs = {
            'ignore_missing': False,
            'domain_id': None,
        }
        self.identity_sdk_client.find_project.assert_called_once_with(
            name_or_id=self.project.id, **kwargs
        )

        self.assertIsNone(result)

    def test_user_set_project_domain(self):
        arglist = [
            '--project',
            self.project.id,
            '--project-domain',
            self.project.domain_id,
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('project', self.project.id),
            ('project_domain', self.project.domain_id),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
            'default_project_id': self.project.id,
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )

        self.identity_sdk_client.find_domain.assert_called_once_with(
            name_or_id=self.project.domain_id, ignore_missing=False
        )

        self.assertIsNone(result)

    def test_user_set_enable(self):
        arglist = [
            '--enable',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('project', None),
            ('enable', True),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_disable(self):
        arglist = [
            '--disable',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', True),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': False,
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_ignore_lockout_failure_attempts(self):
        arglist = [
            '--ignore-lockout-failure-attempts',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('ignore_lockout_failure_attempts', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'ignore_lockout_failure_attempts': True},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_no_ignore_lockout_failure_attempts(self):
        arglist = [
            '--no-ignore-lockout-failure-attempts',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('no_ignore_lockout_failure_attempts', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'ignore_lockout_failure_attempts': False},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_ignore_password_expiry(self):
        arglist = [
            '--ignore-password-expiry',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('ignore_password_expiry', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'ignore_password_expiry': True},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_no_ignore_password_expiry(self):
        arglist = [
            '--no-ignore-password-expiry',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('no_ignore_password_expiry', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'ignore_password_expiry': False},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_ignore_change_password_upon_first_use(self):
        arglist = [
            '--ignore-change-password-upon-first-use',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('ignore_change_password_upon_first_use', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'ignore_change_password_upon_first_use': True},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_no_ignore_change_password_upon_first_use(self):
        arglist = [
            '--no-ignore-change-password-upon-first-use',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('no_ignore_change_password_upon_first_use', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'ignore_change_password_upon_first_use': False},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_enable_lock_password(self):
        arglist = [
            '--enable-lock-password',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('enable_lock_password', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'lock_password': True},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_disable_lock_password(self):
        arglist = [
            '--disable-lock-password',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('disable_lock_password', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'lock_password': False},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_enable_multi_factor_auth(self):
        arglist = [
            '--enable-multi-factor-auth',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('enable_multi_factor_auth', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'multi_factor_auth_enabled': True},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_disable_multi_factor_auth(self):
        arglist = [
            '--disable-multi-factor-auth',
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('disable_multi_factor_auth', True),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'multi_factor_auth_enabled': False},
        }
        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_option_multi_factor_auth_rule(self):
        arglist = [
            '--multi-factor-auth-rule',
            identity_fakes.mfa_opt1,
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('multi_factor_auth_rules', [identity_fakes.mfa_opt1]),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {'multi_factor_auth_rules': [["password", "totp"]]},
        }

        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)

    def test_user_set_with_multiple_options(self):
        arglist = [
            '--ignore-password-expiry',
            '--enable-multi-factor-auth',
            '--multi-factor-auth-rule',
            identity_fakes.mfa_opt1,
            self.user.name,
        ]
        verifylist = [
            ('name', None),
            ('email', None),
            ('ignore_password_expiry', True),
            ('enable_multi_factor_auth', True),
            ('multi_factor_auth_rules', [identity_fakes.mfa_opt1]),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'is_enabled': True,
            'options': {
                'ignore_password_expiry': True,
                'multi_factor_auth_enabled': True,
                'multi_factor_auth_rules': [["password", "totp"]],
            },
        }

        self.identity_sdk_client.update_user.assert_called_with(
            user=self.user, **kwargs
        )
        self.assertIsNone(result)


class TestUserSetPassword(identity_fakes.TestIdentity):
    def setUp(self):
        super().setUp()
        self.cmd = user.SetPasswordUser(self.app, None)

    @staticmethod
    @contextlib.contextmanager
    def _mock_get_password(*passwords):
        mocker = mock.Mock(side_effect=passwords)
        with mock.patch("osc_lib.utils.get_password", mocker):
            yield

    def test_user_password_change(self):
        current_pass = 'old_pass'
        new_pass = 'new_pass'
        arglist = [
            '--password',
            new_pass,
        ]
        verifylist = [
            ('password', new_pass),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Mock getting user current password.
        with self._mock_get_password(current_pass):
            result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        conn = self.app.client_manager.sdk_connection
        user_id = conn.config.get_auth().get_user_id(conn.identity)

        self.identity_sdk_client.update_password.assert_called_with(
            user=user_id, current_password=current_pass, password=new_pass
        )

    def test_user_create_password_prompt(self):
        current_pass = 'old_pass'
        new_pass = 'new_pass'
        parsed_args = self.check_parser(self.cmd, [], [])

        # Mock getting user current and new password.
        with self._mock_get_password(current_pass, new_pass):
            result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        conn = self.app.client_manager.sdk_connection
        user_id = conn.config.get_auth().get_user_id(conn.identity)

        self.identity_sdk_client.update_password.assert_called_with(
            user=user_id, current_password=current_pass, password=new_pass
        )

    def test_user_password_change_no_prompt(self):
        current_pass = 'old_pass'
        new_pass = 'new_pass'
        arglist = [
            '--password',
            new_pass,
            '--original-password',
            current_pass,
        ]
        verifylist = [
            ('password', new_pass),
            ('original_password', current_pass),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        conn = self.app.client_manager.sdk_connection
        user_id = conn.config.get_auth().get_user_id(conn.identity)

        self.identity_sdk_client.update_password.assert_called_with(
            user=user_id, current_password=current_pass, password=new_pass
        )


class TestUserShow(identity_fakes.TestIdentity):
    user = sdk_fakes.generate_fake_resource(_user.User)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.find_user.return_value = self.user

        # Get the command object to test
        self.cmd = user.ShowUser(self.app, None)
        self.identity_client.auth.client.get_user_id.return_value = (
            self.user.id
        )
        self.identity_client.tokens.get_token_data.return_value = {
            'token': {
                'user': {
                    'domain_id': {'id': self.user.domain_id},
                    'id': self.user.id,
                    'name': self.user.name,
                }
            }
        }

    def test_user_show(self):
        arglist = [
            self.user.id,
        ]
        verifylist = [
            ('user', self.user.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_user.assert_called_with(
            name_or_id=self.user.id, ignore_missing=False
        )

        collist = (
            'default_project_id',
            'domain_id',
            'email',
            'enabled',
            'id',
            'name',
            'description',
            'password_expires_at',
            'options',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.user.default_project_id,
            self.user.domain_id,
            self.user.email,
            True,
            self.user.id,
            self.user.name,
            self.user.description,
            self.user.password_expires_at,
            getattr(self.user, 'options', {}),
        )
        self.assertEqual(datalist, data)

    def test_user_show_with_domain(self):
        user = sdk_fakes.generate_fake_resource(
            resource_type=_user.User, name=self.user.name
        )

        arglist = [
            "--domain",
            self.user.domain_id,
            user.name,
        ]
        verifylist = [
            ('domain', self.user.domain_id),
            ('user', user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        user_str = common._get_token_resource(
            self.identity_sdk_client,
            'user',
            parsed_args.user,
            parsed_args.domain,
        )
        self.assertEqual(self.user.name, user_str)

        arglist = [
            "--domain",
            user.domain_id,
            user.name,
        ]
        verifylist = [
            ('domain', user.domain_id),
            ('user', user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        user_str = common._get_token_resource(
            self.identity_sdk_client,
            'user',
            parsed_args.user,
            parsed_args.domain,
        )
        self.assertEqual(user.name, user_str)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from typing import ClassVar
import uuid

from openstackclient.tests.functional.network.v2 import common


class IPAvailabilityTests(common.NetworkTests):
    """Functional tests for IP availability"""

    NAME: ClassVar[str]
    NETWORK_NAME: ClassVar[str]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        if cls.haz_network:
            cls.NAME = uuid.uuid4().hex
            cls.NETWORK_NAME = uuid.uuid4().hex

            # Create a network for the subnet
            cls.openstack('network create ' + cls.NETWORK_NAME)
            cmd_output = cls.openstack(
                'subnet create '
                + '--network '
                + cls.NETWORK_NAME
                + ' '
                + '--subnet-range 10.10.10.0/24 '
                + cls.NAME,
                parse_output=True,
            )
            cls.assertOutput(cls.NAME, cmd_output['name'])

    @classmethod
    def tearDownClass(cls):
        try:
            if cls.haz_network:
                raw_subnet = cls.openstack('subnet delete ' + cls.NAME)
                raw_network = cls.openstack(
                    'network delete ' + cls.NETWORK_NAME
                )
                cls.assertOutput('', raw_subnet)
                cls.assertOutput('', raw_network)
        finally:
            super().tearDownClass()

    def test_ip_availability_list(self):
        """Test ip availability list"""
        cmd_output = self.openstack(
            'ip availability list',
            parse_output=True,
        )
        names = [x['Network Name'] for x in cmd_output]
        self.assertIn(self.NETWORK_NAME, names)

    def test_ip_availability_show(self):
        """Test ip availability show"""
        cmd_output = self.openstack(
            'ip availability show ' + self.NETWORK_NAME,
            parse_output=True,
        )
        self.assertEqual(
            self.NETWORK_NAME,
            cmd_output['network_name'],
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.block_storage.v3 import attachment as _attachment
from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import envvars
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common

LOG = logging.getLogger(__name__)

_FILTER_DEPRECATED = _(
    "This option is deprecated. Consider using the '--filters' option which "
    "was introduced in microversion 3.33 instead."
)


def _format_attachment(
    attachment: _attachment.Attachment,
) -> tuple[tuple[str, ...], Iterable[Any]]:
    columns = (
        'id',
        'volume_id',
        'instance',
        'status',
        'attach_mode',
        'attached_at',
        'detached_at',
        'connection_info',
    )
    column_headers = (
        'ID',
        'Volume ID',
        'Instance ID',
        'Status',
        'Attach Mode',
        'Attached At',
        'Detached At',
        'Properties',
    )

    # VolumeAttachmentManager.create returns a dict while everything else
    # returns a VolumeAttachment object
    if isinstance(attachment, dict):
        data: tuple[Any, ...] = ()
        for column in columns:
            if column == 'connection_info':
                data += (format_columns.DictColumn(attachment[column]),)
                continue
            data += (attachment[column],)
    else:
        data = utils.get_item_properties(
            attachment,
            columns,
            formatters={
                'connection_info': format_columns.DictColumn,
            },
        )

    # TODO(stephenfin): Improve output with the nested connection_info
    # field - cinderclient printed two things but that's equally ugly
    return (column_headers, data)


class CreateVolumeAttachment(command.ShowOne):
    """Create an attachment for a volume.

    This command will only create a volume attachment in the Volume service. It
    will not invoke the necessary Compute service actions to actually attach
    the volume to the server at the hypervisor level. As a result, it should
    typically only be used for troubleshooting issues with an existing server
    in combination with other tooling. For all other use cases, the 'server
    add volume' command should be preferred.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Name or ID of volume to attach to server.'),
        )
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Name or ID of server to attach volume to.'),
        )
        parser.add_argument(
            '--connect',
            action='store_true',
            dest='connect',
            default=False,
            help=_('Make an active connection using provided connector info'),
        )
        parser.add_argument(
            '--no-connect',
            action='store_false',
            dest='connect',
            help=_(
                'Do not make an active connection using provided connector '
                'info'
            ),
        )
        parser.add_argument(
            '--initiator',
            metavar='<initiator>',
            help=_('IQN of the initiator attaching to'),
        )
        parser.add_argument(
            '--ip',
            metavar='<ip>',
            help=_('IP of the system attaching to'),
        )
        parser.add_argument(
            '--host',
            metavar='<host>',
            help=_('Name of the host attaching to'),
        )
        parser.add_argument(
            '--platform',
            metavar='<platform>',
            help=_('Platform type'),
        )
        parser.add_argument(
            '--os-type',
            metavar='<ostype>',
            help=_('OS type'),
        )
        parser.add_argument(
            '--multipath',
            action='store_true',
            dest='multipath',
            default=False,
            help=_('Use multipath'),
        )
        parser.add_argument(
            '--no-multipath',
            action='store_false',
            dest='multipath',
            help=_('Use multipath'),
        )
        parser.add_argument(
            '--mountpoint',
            metavar='<mountpoint>',
            help=_('Mountpoint volume will be attached at'),
        )
        parser.add_argument(
            '--mode',
            metavar='<mode>',
            help=_(
                'Mode of volume attachment, rw, ro and null, where null '
                'indicates we want to honor any existing admin-metadata '
                'settings '
                '(supported by --os-volume-api-version 3.54 or later)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        compute_client = self.app.client_manager.compute

        if not sdk_utils.supports_microversion(volume_client, '3.27'):
            msg = _(
                "--os-volume-api-version 3.27 or greater is required to "
                "support the 'volume attachment create' command"
            )
            raise exceptions.CommandError(msg)

        if parsed_args.mode:
            if not sdk_utils.supports_microversion(volume_client, '3.54'):
                msg = _(
                    "--os-volume-api-version 3.54 or greater is required to "
                    "support the '--mode' option"
                )
                raise exceptions.CommandError(msg)

        connector = {}
        if parsed_args.connect:
            connector = {
                'initiator': parsed_args.initiator,
                'ip': parsed_args.ip,
                'platform': parsed_args.platform,
                'host': parsed_args.host,
                'os_type': parsed_args.os_type,
                'multipath': parsed_args.multipath,
                'mountpoint': parsed_args.mountpoint,
            }
        else:
            if any(
                {
                    parsed_args.initiator,
                    parsed_args.ip,
                    parsed_args.platform,
                    parsed_args.host,
                    parsed_args.host,
                    parsed_args.multipath,
                    parsed_args.mountpoint,
                }
            ):
                msg = _(
                    'You must specify the --connect option for any of the '
                    'connection-specific options such as --initiator to be '
                    'valid'
                )
                raise exceptions.CommandError(msg)

        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )
        server = compute_client.find_server(
            parsed_args.server, ignore_missing=False
        )

        attachment = volume_client.create_attachment(
            volume.id,
            connector=connector,
            instance=server.id,
            mode=parsed_args.mode,
        )

        return _format_attachment(attachment)


class DeleteVolumeAttachment(command.Command):
    """Delete an attachment for a volume.

    Similarly to the 'volume attachment create' command, this command will only
    delete the volume attachment record in the Volume service. It will not
    invoke the necessary Compute service actions to actually attach the volume
    to the server at the hypervisor level. As a result, it should typically
    only be used for troubleshooting issues with an existing server in
    combination with other tooling. For all other use cases, the 'server volume
    remove' command should be preferred.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'attachment',
            metavar='<attachment>',
            help=_('ID of volume attachment to delete'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.27'):
            msg = _(
                "--os-volume-api-version 3.27 or greater is required to "
                "support the 'volume attachment delete' command"
            )
            raise exceptions.CommandError(msg)

        volume_client.delete_attachment(parsed_args.attachment)


class SetVolumeAttachment(command.ShowOne):
    """Update an attachment for a volume.

    This call is designed to be more of an volume attachment completion than
    anything else. It expects the value of a connector object to notify the
    driver that the volume is going to be connected and where it's being
    connected to.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'attachment',
            metavar='<attachment>',
            help=_('ID of volume attachment.'),
        )
        parser.add_argument(
            '--initiator',
            metavar='<initiator>',
            help=_('IQN of the initiator attaching to'),
        )
        parser.add_argument(
            '--ip',
            metavar='<ip>',
            help=_('IP of the system attaching to'),
        )
        parser.add_argument(
            '--host',
            metavar='<host>',
            help=_('Name of the host attaching to'),
        )
        parser.add_argument(
            '--platform',
            metavar='<platform>',
            help=_('Platform type'),
        )
        parser.add_argument(
            '--os-type',
            metavar='<ostype>',
            help=_('OS type'),
        )
        parser.add_argument(
            '--multipath',
            action='store_true',
            dest='multipath',
            default=False,
            help=_('Use multipath'),
        )
        parser.add_argument(
            '--no-multipath',
            action='store_false',
            dest='multipath',
            help=_('Use multipath'),
        )
        parser.add_argument(
            '--mountpoint',
            metavar='<mountpoint>',
            help=_('Mountpoint volume will be attached at'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.27'):
            msg = _(
                "--os-volume-api-version 3.27 or greater is required to "
                "support the 'volume attachment set' command"
            )
            raise exceptions.CommandError(msg)

        connector = {
            'initiator': parsed_args.initiator,
            'ip': parsed_args.ip,
            'platform': parsed_args.platform,
            'host': parsed_args.host,
            'os_type': parsed_args.os_type,
            'multipath': parsed_args.multipath,
            'mountpoint': parsed_args.mountpoint,
        }

        attachment = volume_client.update_attachment(
            parsed_args.attachment,
            connector=connector,
        )

        return _format_attachment(attachment)


class CompleteVolumeAttachment(command.Command):
    """Complete an attachment for a volume."""

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'attachment',
            metavar='<attachment>',
            help=_('ID of volume attachment to mark as completed'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.44'):
            msg = _(
                "--os-volume-api-version 3.44 or greater is required to "
                "support the 'volume attachment complete' command"
            )
            raise exceptions.CommandError(msg)

        volume_client.complete_attachment(parsed_args.attachment)


class ListVolumeAttachment(command.Lister):
    """Lists all volume attachments."""

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            dest='project',
            metavar='<project>',
            help=_('Filter results by project (name or ID) (admin only)'),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--all-projects',
            dest='all_projects',
            action='store_true',
            default=envvars.boolenv('ALL_PROJECTS'),
            help=_('Shows details for all projects (admin only).'),
        )
        parser.add_argument(
            '--volume-id',
            metavar='<volume-id>',
            default=None,
            help=_('Filters results by a volume ID. ') + _FILTER_DEPRECATED,
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            help=_('Filters results by a status. ') + _FILTER_DEPRECATED,
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        # TODO(stephenfin): Add once we have an equivalent command for
        # 'cinder list-filters'
        # parser.add_argument(
        #     '--filter',
        #     metavar='<key=value>',
        #     action=parseractions.KeyValueAction,
        #     dest='filters',
        #     help=_(
        #         "Filter key and value pairs. Use 'foo' to "
        #         "check enabled filters from server. Use 'key~=value' for "
        #         "inexact filtering if the key supports "
        #         "(supported by --os-volume-api-version 3.33 or above)"
        #     ),
        # )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.27'):
            msg = _(
                "--os-volume-api-version 3.27 or greater is required to "
                "support the 'volume attachment list' command"
            )
            raise exceptions.CommandError(msg)

        project_id = None
        if parsed_args.project:
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )

        search_opts = {
            'all_tenants': True if project_id else parsed_args.all_projects,
            'project_id': project_id,
            'status': parsed_args.status,
            'volume_id': parsed_args.volume_id,
        }
        # Update search option with `filters`
        # if AppendFilters.filters:
        #     search_opts.update(shell_utils.extract_filters(AppendFilters.filters))  # noqa: E501

        # TODO(stephenfin): Implement sorting
        attachments = volume_client.attachments(
            search_opts=search_opts,
            marker=parsed_args.marker,
            limit=parsed_args.limit,
            max_items=parsed_args.max_items,
        )

        column_headers = (
            'ID',
            'Volume ID',
            'Server ID',
            'Status',
        )
        columns = (
            'id',
            'volume_id',
            'instance',
            'status',
        )

        return (
            column_headers,
            (utils.get_item_properties(a, columns) for a in attachments),
        )


class ShowVolumeAttachment(command.ShowOne):
    """Show detailed information for a volume attachment."""

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'attachment',
            metavar='<attachment>',
            help=_('ID of volume attachment.'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.27'):
            msg = _(
                "--os-volume-api-version 3.27 or greater is required to "
                "support the 'volume attachment show' command"
            )
            raise exceptions.CommandError(msg)

        attachment = volume_client.get_attachment(parsed_args.attachment)

        return _format_attachment(attachment)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Identity v3 EC2 Credentials action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any, cast

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


def _determine_ec2_user(
    parsed_args: argparse.Namespace, client_manager: Any
) -> str:
    """Determine a user several different ways.

    Assumes parsed_args has user and user_domain arguments. Attempts to find
    the user if domain scoping is provided, otherwise revert to a basic user
    call. Lastly use the currently authenticated user.

    """

    user_domain = None
    if parsed_args.user_domain:
        user_domain = common.find_domain(
            client_manager.identity, parsed_args.user_domain
        )
    if parsed_args.user:
        if user_domain is not None:
            user = utils.find_resource(
                client_manager.identity.users,
                parsed_args.user,
                domain_id=user_domain.id,
            ).id
        else:
            user = utils.find_resource(
                client_manager.identity.users, parsed_args.user
            ).id
    else:
        # Get the user from the current auth
        user = client_manager.auth_ref.user_id
    return cast(str, user)


class CreateEC2Creds(command.ShowOne):
    _description = _("Create EC2 credentials")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                'Create credentials in project '
                '(name or ID; default: current authenticated project)'
            ),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_(
                'Create credentials for user '
                '(name or ID; default: current authenticated user)'
            ),
        )
        common.add_user_domain_option_to_parser(parser)
        common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = self.app.client_manager.identity
        client_manager = self.app.client_manager
        user = _determine_ec2_user(parsed_args, client_manager)

        project_domain = None
        if parsed_args.project_domain:
            project_domain = common.find_domain(
                identity_client, parsed_args.project_domain
            )

        if parsed_args.project:
            if project_domain is not None:
                project = utils.find_resource(
                    identity_client.projects,
                    parsed_args.project,
                    domain_id=project_domain.id,
                ).id
            else:
                project = utils.find_resource(
                    identity_client.projects, parsed_args.project
                ).id
        else:
            # Get the project from the current auth
            project = self.app.client_manager.auth_ref.project_id

        creds = identity_client.ec2.create(user, project)

        info = {}
        info.update(creds._info)

        if 'tenant_id' in info:
            info.update({'project_id': info.pop('tenant_id')})

        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data


class DeleteEC2Creds(command.Command):
    _description = _("Delete EC2 credentials")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'access_key',
            metavar='<access-key>',
            nargs='+',
            help=_('Credentials access key(s)'),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Delete credentials for user (name or ID)'),
        )
        common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client_manager = self.app.client_manager
        user = _determine_ec2_user(parsed_args, client_manager)
        result = 0
        for i in parsed_args.access_key:
            try:
                client_manager.identity.ec2.delete(user, i)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete EC2 credentials with "
                        "access key '%(access_key)s': %(e)s"
                    ),
                    {'access_key': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.access_key)
            msg = _("%(result)s of %(total)s EC2 keys failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListEC2Creds(command.Lister):
    _description = _("List EC2 credentials")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Filter list by user (name or ID)'),
        )
        common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client_manager = self.app.client_manager
        user = _determine_ec2_user(parsed_args, client_manager)

        columns = ('access', 'secret', 'tenant_id', 'user_id')
        column_headers = ('Access', 'Secret', 'Project ID', 'User ID')
        data = client_manager.identity.ec2.list(user)

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class ShowEC2Creds(command.ShowOne):
    _description = _("Display EC2 credentials details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'access_key',
            metavar='<access-key>',
            help=_('Credentials access key'),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Show credentials for user (name or ID)'),
        )
        common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client_manager = self.app.client_manager
        user = _determine_ec2_user(parsed_args, client_manager)
        creds = client_manager.identity.ec2.get(user, parsed_args.access_key)

        info = {}
        info.update(creds._info)

        if 'tenant_id' in info:
            info.update({'project_id': info.pop('tenant_id')})

        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from openstack.identity.v3 import project as _project
from openstack.test import fakes as sdk_fakes

from openstackclient.common import project_cleanup
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit import utils as test_utils


class TestProjectCleanup(
    identity_fakes.FakeClientMixin, test_utils.TestCommand
):
    def setUp(self):
        super().setUp()
        self.cmd = project_cleanup.ProjectCleanup(self.app, None)

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project
        self.app.client_manager.sdk_connection.connect_as_project.return_value = self.app.client_manager.sdk_connection

    def test_project_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_project_cleanup_with_filters(self):
        arglist = [
            '--project',
            self.project.id,
            '--created-before',
            '2200-01-01',
            '--updated-before',
            '2200-01-02',
        ]
        verifylist = [
            ('dry_run', False),
            ('auth_project', False),
            ('project', self.project.id),
            ('created_before', '2200-01-01'),
            ('updated_before', '2200-01-02'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = None

        with mock.patch('getpass.getpass', return_value='y'):
            result = self.cmd.take_action(parsed_args)

        self.app.client_manager.sdk_connection.connect_as_project.assert_called_with(
            self.project
        )
        filters = {'created_at': '2200-01-01', 'updated_at': '2200-01-02'}

        calls = [
            mock.call(
                dry_run=True,
                status_queue=mock.ANY,
                filters=filters,
                skip_resources=None,
            ),
            mock.call(
                dry_run=False,
                status_queue=mock.ANY,
                filters=filters,
                skip_resources=None,
            ),
        ]
        self.app.client_manager.sdk_connection.project_cleanup.assert_has_calls(
            calls
        )

        self.assertIsNone(result)

    def test_project_cleanup_with_auto_approve(self):
        arglist = [
            '--project',
            self.project.id,
            '--auto-approve',
        ]
        verifylist = [
            ('dry_run', False),
            ('auth_project', False),
            ('project', self.project.id),
            ('auto_approve', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = None

        result = self.cmd.take_action(parsed_args)

        self.app.client_manager.sdk_connection.connect_as_project.assert_called_with(
            self.project
        )
        calls = [
            mock.call(
                dry_run=True,
                status_queue=mock.ANY,
                filters={},
                skip_resources=None,
            ),
            mock.call(
                dry_run=False,
                status_queue=mock.ANY,
                filters={},
                skip_resources=None,
            ),
        ]
        self.app.client_manager.sdk_connection.project_cleanup.assert_has_calls(
            calls
        )

        self.assertIsNone(result)

    def test_project_cleanup_with_project(self):
        arglist = [
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('dry_run', False),
            ('auth_project', False),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = None

        with mock.patch('getpass.getpass', return_value='y'):
            result = self.cmd.take_action(parsed_args)

        self.app.client_manager.sdk_connection.connect_as_project.assert_called_with(
            self.project
        )
        calls = [
            mock.call(
                dry_run=True,
                status_queue=mock.ANY,
                filters={},
                skip_resources=None,
            ),
            mock.call(
                dry_run=False,
                status_queue=mock.ANY,
                filters={},
                skip_resources=None,
            ),
        ]
        self.app.client_manager.sdk_connection.project_cleanup.assert_has_calls(
            calls
        )

        self.assertIsNone(result)

    def test_project_cleanup_with_project_abort(self):
        arglist = [
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('dry_run', False),
            ('auth_project', False),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = None

        with mock.patch('getpass.getpass', return_value='y'):
            result = self.cmd.take_action(parsed_args)

        self.app.client_manager.sdk_connection.connect_as_project.assert_called_with(
            self.project
        )
        calls = [
            mock.call(
                dry_run=True,
                status_queue=mock.ANY,
                filters={},
                skip_resources=None,
            ),
        ]
        self.app.client_manager.sdk_connection.project_cleanup.assert_has_calls(
            calls
        )

        self.assertIsNone(result)

    def test_project_cleanup_with_dry_run(self):
        arglist = [
            '--dry-run',
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('dry_run', True),
            ('auth_project', False),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = None

        result = self.cmd.take_action(parsed_args)

        self.app.client_manager.sdk_connection.connect_as_project.assert_called_with(
            self.project
        )
        self.app.client_manager.sdk_connection.project_cleanup.assert_called_once_with(
            dry_run=True,
            status_queue=mock.ANY,
            filters={},
            skip_resources=None,
        )

        self.assertIsNone(result)

    def test_project_cleanup_with_auth_project(self):
        self.app.client_manager.auth_ref = mock.Mock()
        self.app.client_manager.auth_ref.project_id = self.project.id
        arglist = [
            '--auth-project',
        ]
        verifylist = [
            ('dry_run', False),
            ('auth_project', True),
            ('project', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = None

        with mock.patch('getpass.getpass', return_value='y'):
            result = self.cmd.take_action(parsed_args)

        self.app.client_manager.sdk_connection.connect_as_project.assert_not_called()
        calls = [
            mock.call(
                dry_run=True,
                status_queue=mock.ANY,
                filters={},
                skip_resources=None,
            ),
            mock.call(
                dry_run=False,
                status_queue=mock.ANY,
                filters={},
                skip_resources=None,
            ),
        ]
        self.app.client_manager.sdk_connection.project_cleanup.assert_has_calls(
            calls
        )

        self.assertIsNone(result)

    def test_project_cleanup_with_skip_resource(self):
        skip_resource = "block_storage.backup"
        arglist = [
            '--project',
            self.project.id,
            '--skip-resource',
            skip_resource,
        ]
        verifylist = [('skip_resource', [skip_resource])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = None

        with mock.patch('getpass.getpass', return_value='y'):
            result = self.cmd.take_action(parsed_args)

        self.app.client_manager.sdk_connection.connect_as_project.assert_called_with(
            self.project
        )

        calls = [
            mock.call(
                dry_run=True,
                status_queue=mock.ANY,
                filters={},
                skip_resources=[skip_resource],
            ),
            mock.call(
                dry_run=False,
                status_queue=mock.ANY,
                filters={},
                skip_resources=[skip_resource],
            ),
        ]
        self.app.client_manager.sdk_connection.project_cleanup.assert_has_calls(
            calls
        )

        self.assertIsNone(result)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional.image import base


class ImageMetadefResourceTypeTests(base.BaseImageTests):
    """Functional tests for image metadef resource type commands."""

    def setUp(self):
        super().setUp()

        # Create unique namespace name using UUID
        self.namespace_name = 'test-mdef-ns-' + uuid.uuid4().hex
        self.resource_type_name = 'test-mdef-rt-' + uuid.uuid4().hex

        # Create namespace
        self.openstack('image metadef namespace create ' + self.namespace_name)
        self.addCleanup(
            self.openstack,
            'image metadef namespace delete ' + self.namespace_name,
        )

    def test_metadef_resource_type(self):
        """Test image metadef resource type commands"""

        self.openstack(
            'image metadef resource type association create '
            f'{self.namespace_name} {self.resource_type_name}',
        )
        self.addCleanup(
            self.openstack,
            'image metadef resource type association delete '
            f'{self.namespace_name} {self.resource_type_name}',
        )

        output = self.openstack(
            'image metadef resource type list',
            parse_output=True,
        )

        self.assertIn(
            self.resource_type_name, [item['Name'] for item in output]
        )
